"""
This module provides a command line client for the aptdaemon
"""
# Copyright (C) 2008-2009 Sebastian Heinlein <sevel@glatzor.de>
#
# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

__author__ = "Sebastian Heinlein <devel@glatzor.de>"

import array
import fcntl
from gettext import gettext as _
from optparse import OptionParser
import os
import pty
import termios
import tty
import signal
import sys

import apt_pkg
import gobject
import dbus.glib

import aptdaemon
import client
import enums
from misc import threaded

ANSI_BOLD = chr(27) + "[1m"
ANSI_RESET = chr(27) + "[0m"


class ConsoleClient:
    """
    Command line interface client to aptdaemon
    """
    def __init__(self, show_terminal=True):
        self._client = client.AptClient()
        self.master_fd, self.slave_fd = pty.openpty()
        self._signals = []
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGQUIT, signal.SIG_IGN)
        signal.signal(signal.SIGWINCH, self._on_terminal_resize)
        self._terminal_width = self._get_terminal_width()
        self._watchers = []
        self._old_tty_mode = None
        self._show_status = True
        self._status = ""
        self._percent = 0
        self._show_terminal = show_terminal
        self._show_progress = True
        self._status_details = ""
        self._progress_details = ""

    def install_packages(self, packages):
        """Install packages"""
        trans = self._client.install_packages(packages,
                                              exit_handler=self._on_exit)
        self._set_transaction(trans)
        trans.run(block=True)

    def install_file(self, path):
        """Install package file."""
        trans = self._client.install_file(path, exit_handler=self._on_exit)
        self._set_transaction(trans)
        trans.run(block=True)

    def remove_packages(self, packages):
        """Remove packages"""
        trans = self._client.remove_packages(packages,
                                             exit_handler=self._on_exit)
        self._set_transaction(trans)
        trans.run(block=True)

    def commit_packages(self, install, reinstall, remove, purge, upgrade):
        """Commit changes"""
        trans = self._client.commit_packages(install, reinstall, remove,
                                             purge, upgrade,
                                             exit_handler=self._on_exit)
        self._set_transaction(trans)
        trans.run(block=True)

    def update_cache(self):
        """Update cache"""
        trans = self._client.update_cache(exit_handler=self._on_exit)
        self._set_transaction(trans)
        trans.run(block=True)

    def upgrade_system(self):
        """Upgrade system"""
        trans = self._client.upgrade_system(exit_handler=self._on_exit)
        self._set_transaction(trans)
        trans.run(block=True)

    def _set_transaction(self, transaction):
        """Monitor the given transaction"""
        for handler in self._signals:
            gobject.remove_source(handler)
        self._signals = []
        self._signals.append(
            transaction.connect("allow-terminal", self._on_allow_terminal))
        self._signals.append(
            transaction.connect("status", self._on_status))
        self._signals.append(
            transaction.connect("status-details", self._on_status_details))
        self._signals.append(
            transaction.connect("progress", self._on_progress))
        self._signals.append(
            transaction.connect("progress-details", self._on_progress_details))
        if self._show_terminal:
            transaction.set_terminal(os.ttyname(self.slave_fd))

    def _on_exit(self, trans, enum):
        """Callback for the exit state of the transaction"""
        # Make sure to dettach the terminal
        self._detach()
        if self._show_progress:
            sys.stderr.write("100% " + ANSI_BOLD +
                             "%-*.*s" % (self._terminal_width,
                                         self._terminal_width,
                                         enums.get_exit_string_from_enum(enum)) +
                             ANSI_RESET + "\n")

        if enum == enums.EXIT_FAILED:
            excep = trans.get_error()
            msg = "%s: %s\n%s\n\n%s" % (
                   _("ERROR"),
                   enums.get_error_string_from_enum(excep.code),
                   enums.get_error_description_from_enum(excep.code),
                   excep.details)
            print msg

    def _on_allow_terminal(self, transaction, allow_terminal):
        """Callback for the AllowTerminal signal of the transaction"""
        if self._show_terminal and allow_terminal == True and \
           not self._watchers:
            sys.stderr.write("\r")
            self._show_progress = False
            self._attach()
        elif allow_terminal == False:
            self._show_progress = True
            self._detach()

    def _on_status(self, transaction, status):
        """Callback for the Status signal of the transaction"""
        self._status = enums.get_status_string_from_enum(status)
        self._update_progress()

    def _on_status_details(self, transaction, text):
        """Callback for the StatusDetails signal of the transaction."""
        self._status_details = text
        self._update_progress()

    def _on_progress_details(self, transaction, items_done, items_total,
                             bytes_done, bytes_total, speed, eta):
        """Callback for the ProgressDetails signal of the transaction."""
        if bytes_total and speed:
            self._progress_details = (_("Downloaded %sB of %sB at %sB/s") % \
                                      (apt_pkg.SizeToStr(bytes_done),
                                       apt_pkg.SizeToStr(bytes_total),
                                       apt_pkg.SizeToStr(speed)))
        elif bytes_total:
            self._progress_details = (_("Downloaded %sB of %sB") % \
                                      (apt_pkg.SizeToStr(bytes_done),
                                       apt_pkg.SizeToStr(bytes_total)))
        else:
            self._progress_details = ""
        self._update_progress()

    def _on_progress(self, transaction, percent):
        """Callback for the Progress signal of the transaction"""
        self._percent = percent
        self._update_progress()

    def _update_progress(self):
        """Update the progress bar."""
        if not self._show_progress:
            return
        text = ANSI_BOLD + self._status + ANSI_RESET
        if self._status_details:
            text += " " + self._status_details
        if self._progress_details:
            text += " (%s)" % self._progress_details
        text_width = self._terminal_width - 5
        sys.stderr.write("%3.3s%% " % self._percent +
                         "%-*.*s" % (text_width, text_width, text) + "\r")

    def _on_terminal_resize(self, signum, frame):
        """Callback for a changed terminal size."""
        self._terminal_width = self._get_terminal_width()
        self._update_progress()

    def _detach(self):
        """Dettach the controlling terminal to aptdaemon."""
        for wid in self._watchers:
            gobject.source_remove(wid)
        if self._old_tty_mode:
            tty.tcsetattr(pty.STDIN_FILENO, tty.TCSAFLUSH,
                          self._old_tty_mode)

    def _attach(self):
        """Attach the controlling terminal to aptdaemon. Based on pty.spwan()"""
        try:
            self._old_tty_mode = tty.tcgetattr(pty.STDIN_FILENO)
            tty.setraw(pty.STDIN_FILENO)
        except tty.error:    # This is the same as termios.error
            self._old_tty_mode = None
        flags = gobject.IO_IN | gobject.IO_ERR | gobject.IO_HUP
        wid = gobject.io_add_watch(pty.STDIN_FILENO, flags,
                                   self._copy_io, self.master_fd)
        self._watchers.append(wid)
        wid = gobject.io_add_watch(self.master_fd, flags,
                                   self._copy_io, pty.STDOUT_FILENO)
        self._watchers.append(wid)

    def _copy_io(self, source, condition, target):
        """Callback to copy data between terminals."""
        if condition == gobject.IO_IN:
            data = os.read(source, 1024)
            if target:
                os.write(target, data)
            return True
        os.close(source)
        return False

    def _get_terminal_width(self):
        """Return the witdh in characters of the current terminal."""
        width = array.array("h", fcntl.ioctl(sys.stderr, termios.TIOCGWINSZ,
                                             "\0" * 8))[1]
        return width


def main():
    """Run a command line client for aptdaemon"""
    parser = OptionParser(version=aptdaemon.__version__)
    parser.add_option("-i", "--install", default="",
                      action="store", type="string", dest="install",
                      help=_("Install the given packages"))
    parser.add_option("", "--reinstall", default="",
                      action="store", type="string", dest="reinstall",
                      help=_("Reinstall the given packages"))
    parser.add_option("-r", "--remove", default="",
                      action="store", type="string", dest="remove",
                      help=_("Install the given packages"))
    parser.add_option("-p", "--purge", default="",
                      action="store", type="string", dest="purge",
                      help=_("Remove the given packages including "
                             "configuration files"))
    parser.add_option("-u", "--upgrade", default="",
                      action="store", type="string", dest="upgrade",
                      help=_("Install the given packages"))
    parser.add_option("", "--upgrade-system", default="",
                      action="store", type="string", dest="dist_upgrade",
                      help=_("Upgrade the system"))
    parser.add_option("", "--hide-terminal",
                      action="store_true", dest="hide_terminal",
                      help=_("Do not attach to the apt terminal"))
    (options, args) = parser.parse_args()
    client = ConsoleClient(show_terminal=not options.hide_terminal)
    if options.dist_upgrade:
        client.upgrade_system()
    elif options.install or options.reinstall or options.remove or \
         options.purge or options.remove:
        client.commit_packages(options.install.split(),
                               options.reinstall.split(),
                               options.remove.split(),
                               options.purge.split(),
                               options.upgrade.split())
    else:
        parser.print_help()
        sys.exit(1)

if __name__ == "__main__":
    main()
