#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Core components of aptdaemon.

This module provides the following core classes of the aptdaemon:
AptDaemon - complete daemon for managing software via DBus interface
Transaction - represents a software management operation
TransactionQueue - queue for aptdaemon transactions
Message - non-critical message to the frontend

The main function allows to run the daemon as a command.
"""
# Copyright (C) 2008-2009 Sebastian Heinlein <devel@glatzor.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

__author__  = "Sebastian Heinlein <devel@glatzor.de>"

import collections
import gettext
from gettext import gettext as _
from locale import _parse_localename as parse_localename
import logging
import logging.handlers
from optparse import OptionParser
import os
import Queue
import signal
import sys
import time
import threading
import uuid

import gobject
import dbus.exceptions
import dbus.service
import dbus.mainloop.glib
import dbus.glib

import errors
from enums import *
import policykit
from worker import AptWorker
from misc import get_main_loop

# Setup i18n
gettext.textdomain("aptdaemon")

APTDAEMON_DBUS_INTERFACE = 'org.debian.apt'
APTDAEMON_DBUS_PATH = '/org/debian/apt'
APTDAEMON_DBUS_SERVICE = 'org.debian.apt'

APTDAEMON_TRANSACTION_DBUS_INTERFACE = 'org.debian.apt.transaction'

INACTIVITY_CHECK_INTERVAL = 60
INACTIVITY_TIMEOUT = 5 * 60

# Setup the DBus main loop
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

# Required for daemon mode
os.putenv("PATH",
          "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin")

# Setup logging to syslog and the console
_syslog_handler = logging.handlers.SysLogHandler(address="/dev/log",
                             facility=logging.handlers.SysLogHandler.LOG_DAEMON)
_syslog_formatter = logging.Formatter("%(name)s: %(levelname)s: %(message)s")
_syslog_handler.setFormatter(_syslog_formatter)
_console_handler = logging.StreamHandler()
_console_formatter = logging.Formatter("%(asctime)s %(name)s [%(levelname)s]: "
                                       "%(message)s",
                                       "%T")
_console_handler.setFormatter(_console_formatter)
log = logging.getLogger("AptDaemon")
log.addHandler(_console_handler)
log.addHandler(_syslog_handler)
#FIXME: Use LoggerAdapter (requires Python 2.6)
log_trans = logging.getLogger("AptDaemon.Trans")


class Message:
    """Non-critical message from the package manager to the user."""
    def __init__(self, enum, details=""):
        self.enum = enum
        self.details = details

class Transaction(dbus.service.Object):

    """Represents a transaction on the D-Bus.

    A transaction represents a single package management task, e.g. installation
    or removal of packages. This class allows to expose information and to
    controll the transaction via DBus using policykit for managing
    privileges.
    """

    def __init__(self, queue, caller, connect=True, bus=None):
        """Initialize a new Transaction instance.

        Keyword arguments:
        queue -- TransactionQueue instance of the daemon
        caller -- DBus caller who requested the creation of the Transaction
        connect -- if the Transaction should connect to DBus (default is True)
        bus -- the DBus connection which should be used (defaults to system bus)
        """
        id = uuid.uuid4().get_hex()
        self.tid = "/org/debian/apt/transaction/%s" % id
        if connect == True:
            if bus is None:
                bus = dbus.SystemBus()
            bus_name = dbus.service.BusName(APTDAEMON_DBUS_SERVICE, bus)
            dbus_path = self.tid
        else:
            bus_name = None
            dbus_path = None
        dbus.service.Object.__init__(self, bus_name, dbus_path)
        self.queue = queue
        self.caller = caller
        self.locale = None
        self.terminal = None
        self.debconf = None
        self.kwargs = None
        self._role = ROLE_UNSET
        self._progress = 0
        self._progress_details = (0, 0, 0, 0, 0, 0)
        self._exit = None
        self._status = STATUS_SETTING_UP
        self._status_details = ""
        self._error = None
        self._allow_cancel = True
        self._allow_terminal = False
        self._required_medium = None
        self.cancelled = False
        self.paused = False
        self.messages = []

    def _set_role(self, enum):
        self._role = enum
        self.Role(enum)

    def _get_role(self):
        return self._role

    role = property(_get_role, _set_role, doc="Operation type of transaction.")

    def _set_progress_details(self, details):
        items_done, items_total, bytes_done, bytes_total, speed, eta = details
        self._set_progress_details = details
        self.ProgressDetails(items_done, items_total, bytes_done, bytes_total, speed, eta)

    def _get_progress_details(self):
        return self._progress_details

    progress_details = property(_get_progress_details, _set_progress_details,
                                doc = "Tuple containing detailed progress information: "
                                       "items done, total items, bytes done, total "
                                       "bytes, speed and remaining time")

    def _set_error(self, excep):
        self._error = excep
        self.ErrorCode(excep.enum, excep.message)

    def _get_error(self):
        return self._error

    error = property(_get_error, _set_error, doc="Raised exception.")

    def _set_exit(self, enum):
        self._exit = enum
        self.Finished(enum)
        self.status = STATUS_FINISHED

    def _get_exit(self):
        return self._exit

    exit = property(_get_exit, _set_exit, doc="The exit state of the transaction.")

    def _get_status(self):
        return self._status

    def _set_status(self, enum):
        self._status = enum
        self.Status(enum)

    status = property(_get_status, _set_status, doc="The status of the transaction.")

    def _get_status_details(self):
        return self._status_details

    def _set_status_details(self, text):
        self._status_details = text
        self.StatusDetails(text)

    status_details = property(_get_status_details, _set_status_details,
                              doc="The status message from apt.")

    def _get_progress(self):
        return self._progress

    def _set_progress(self, percent):
        self._progress = percent
        self.Progress(percent)

    progress = property(_get_progress, _set_progress,
                      doc="The progress of the transaction in percent.")

    def _get_allow_cancel(self):
        return self._allow_cancel

    def _set_allow_cancel(self, allow_cancel):
        self._allow_cancel = allow_cancel
        self.AllowCancel(allow_cancel)

    allow_cancel = property(_get_allow_cancel, _set_allow_cancel,
                            doc="If it's currently allowed to cancel the transaction.")

    def _get_allow_terminal(self):
        return self._allow_terminal

    def _set_allow_terminal(self, allow_terminal):
        self._allow_terminal = allow_terminal
        self.AllowTerminal(allow_terminal)

    allow_terminal = property(_get_allow_terminal, _set_allow_terminal,
                              doc="If it's currently allowed to attach a terminal "
                                  "to the transaction.")

    def _get_required_medium(self):
        return self._required_medium

    def _set_required_medium(self, medium):
        label, drive = medium
        self._required_medium = medium
        self.MediumRequired(label, drive)

    required_medium = property(_get_required_medium, _set_required_medium,
                               doc="Tuple containing the label and the drive of a "
                                   "required CD/DVD to install packages from.")

    # Signals

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="i")
    def Role(self, enum):
        """Set and emit the transaction role.

        The role of the transaction specifies the type of operation type,
        e.g. installation, removal or upgrade.

        Keyword argument:
        enum -- the type of the transaction e.g. ROLE_UPDATE_CACHE
        """
        log_trans.info("Emitting Role: %s" % \
                       get_role_localised_present_from_enum(enum))

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="is")
    def ErrorCode(self, error, details):
        """Set and emit the error code to provide information about a failed 
        transaction.

        Keyword arguments:
        error -- the error type, e.g. ERROR_NO_CACHE
        details -- optional clear text information about the error
        """
        log_trans.info("Emitting ErrorCode: %s - %s" % \
                       (get_error_string_from_enum(error), details))

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="is")
    def Message(self, enum, details):
        """Store and emit message about a non-critical issue.

        Keyword arguments:
        enum -- the message type, e.g. MESSAGE_SYSTEM_NOT_UPTODATE
        """
        log_trans.info("Emitting Message: %s - %s" % \
                       (get_msg_string_from_enum(enum), details))

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="i")
    def Finished(self, enum):
        """Mark and emit the transaction as finished.

        Keyword argument:
        enum -- the exit state of the transaction, e.g. EXIT_FAILED
        """
        log_trans.info("Emitting Finished: %s" % get_exit_string_from_enum(exit))

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="i")
    def Status(self, enum):
        """Set and emit the the status of the transaction.

        Keyword argument:
        enum -- the status type, e.g. STATUS_DOWNLOADING
        """
        log_trans.info("Emitting Status: %s" % \
                       get_status_string_from_enum(enum))

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="s")
    def StatusDetails(self, text):
        """Emit clear text description of the current transaction status.

        Keyword argument:
        text -- clear text description of the status
        """
        log_trans.info("Emitting StatusDetails: %s" % text)

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="i")
    def Progress(self, percentage):
        """Set and emit the the progress of the transaction in percent.

        Keyword argument:
        percentage -- progress in percent, ranges from 0 to 100
        """
        log_trans.info("Emitting Progress: %s" % percentage)

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="iiiiii")
    def ProgressDetails(self, items_done, items_total, bytes_done, bytes_total,
                        speed, eta):
        """Emit detailed information about the transaction process.

        Keyword arguments:
        items_done -- the number of already processed items
        items_total -- the total number of items to process
        bytes_done -- the total number of already processed bytes
        bytes_total -- the total number of bytes to process
        speed -- the current download speed in cps
        eta -- the remaining time until the transaction will be completed
        """
        log_trans.info("Emitting ProgressDetails: %s of %s items, %s of %s "
                       "bytes at %s cps, eta %s" % (items_done, items_total,
                                                    bytes_done, bytes_total,
                                                    speed, eta))

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="b")
    def AllowCancel(self, allow_cancel):
        """Set and emit if the transaction can currently be cancelled.

        Keyword argument:
        allow_cancel: boolean value if the transaction can be canncelled
        """
        log_trans.info("Emitting AllowCancel: %s" % allow_cancel)

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="b")
    def AllowTerminal(self, allow_terminal):
        """Set and emit if the controlling terminal is attached to the
        worker child process.

        Keyword argument:
        allow_terminal: boolean value if the controlling terminal is attached
        """
        log_trans.info("Emitting AllowTerminal: %s" % allow_terminal)

    @dbus.service.signal(dbus_interface=APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         signature="ss")
    def MediumRequired(self, medium, drive):
        """Set and emit the required medium change.

        This method/signal should be used to inform the user to
        insert the installation CD/DVD:

        Keyword arguments:
        medium -- the CD/DVD label
        drive -- mount point of the drive
        """
        log_trans.info("Emitting MediumRequierd: %s, %s" % (medium, drive))

    # Methods

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="", out_signature="",
                         sender_keyword="sender")
    def UpdateCache(self, sender):
        """Update the package cache.

        Download the latest information about available packages from the
        repositories.

        Keyword argument:
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("UpdateCache() was called")
        self._check_caller(sender)
        if self._role != ROLE_UNSET:
            raise TransactionRoleAlreadySet()
        action = policykit.PK_ACTION_UPDATE_CACHE
        if not policykit.is_authorized_sender(action, sender):
            raise errors.NotAuthorizedError(action, sender)
        self.role = ROLE_UPDATE_CACHE
        self.queue.put(self)

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="as", out_signature="",
                         sender_keyword="sender")
    def RemovePackages(self, package_names, sender):
        """Remove the given packages.

        Keyword arguments:
        package_names -- list of package names to remove
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("RemovePackages() was called: %s" % package_names)
        self._check_caller(sender)
        if self.role != ROLE_UNSET:
            raise TransactionRoleAlreadySet()
        action = policykit.PK_ACTION_REMOVE_PACKAGES
        if not policykit.is_authorized_sender(action, sender):
            raise errors.NotAuthorizedError(action, sender)
        self.role = ROLE_REMOVE_PACKAGES
        self.kwargs = {"package_names" : package_names}
        self.queue.put(self)
        log_trans.debug("Queued RemovePackages()")

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="b", out_signature="",
                         sender_keyword="sender")
    def UpgradeSystem(self, safe_mode, sender):
        """Upgrade the whole system.

        If in safe mode only already installed packages will be updated.
        Updates which require to remove installed packages or to install
        additional packages will be skipped.

        Keyword arguments:
        safe_mode -- boolean value if the safe mode should be used
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("UpgradeSystem() was called with safe mode: "
                       "%s" % safe_mode)
        self._check_caller(sender)
        if self.role != ROLE_UNSET:
            raise TransactionRoleAlreadySet()
        action = policykit.PK_ACTION_UPGRADE_SYSTEM
        if not policykit.is_authorized_sender(action, sender):
            raise errors.NotAuthorizedError(action, sender)
        self.role = ROLE_UPGRADE_SYSTEM
        self.kwargs = {"safe_mode" : safe_mode}
        self.queue.put(self)
        log_trans.debug("Queued UpgradeSystem()")

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="asasasasas", out_signature="",
                         sender_keyword="sender")
    def CommitPackages(self, install, reinstall, remove, purge, upgrade,
                       sender):
        """Perform a complex package operation.

        Keyword arguments:
        install -- package names to be installed
        reinstall -- packages names to be reinstalled
        remove -- package names to be removed
        purge -- package names to be removed including configuration
        upgrade -- package names to be upgraded
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        #FIXME: take sha1 or md5 cash into accout to allow selecting a version
        #       or an origin different from the candidate
        log_trans.info("CommitPackages() was called: %s, %s, %s, %s, %s",
                       install, reinstall, remove, purge, upgrade)
        self._check_caller(sender)
        if self.role != ROLE_UNSET:
            raise TransactionRoleAlreadySet()
        #FIXME: Should we really ask for permission so often?
        if install or reinstall:
            action = policykit.PK_ACTION_INSTALL_PACKAGES
            if not policykit.is_authorized_sender(action, sender):
                raise errors.NotAuthorizedError(action, sender)
        if remove or purge:
            action = policykit.PK_ACTION_REMOVE_PACKAGES
            if not policykit.is_authorized_sender(action, sender):
                raise errors.NotAuthorizedError(action, sender)
        if upgrade:
            action = policykit.PK_ACTION_UPGRADE_PACKAGES
            if not policykit.is_authorized_sender(action, sender):
                raise errors.NotAuthorizedError(action, sender)
        self.role = ROLE_COMMIT_PACKAGES
        self.kwargs = {"install":install, "reinstall":reinstall,
                       "remove":remove, "purge":purge, "upgrade":upgrade}
        log_trans.debug("Queue transaction")
        self.queue.put(self)
        log_trans.debug("Queued")

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="as", out_signature="",
                         sender_keyword="sender")
    def InstallPackages(self, package_names, sender):
        """Install the given packages.

        Keyword arguments:
        package_names -- list of package names to install
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("InstallPackages() was called: %s" % package_names)
        self._check_caller(sender)
        if self.role != ROLE_UNSET:
            raise TransactionRoleAlreadySet()
        action = policykit.PK_ACTION_INSTALL_PACKAGES
        if not policykit.is_authorized_sender(action, sender):
            raise errors.NotAuthorizedError(action, sender)
        self.role = ROLE_INSTALL_PACKAGES
        self.kwargs = {"package_names" : package_names}
        log_trans.debug("Queue transaction")
        self.queue.put(self)
        log_trans.debug("Queued")

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="as", out_signature="",
                         sender_keyword="sender")
    def UpgradePackages(self, package_names, sender):
        """Upgrade the given packages to their latest version.

        Keyword arguments:
        package_names -- list of package names to upgrade
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("UpgradePackages() was called: %s" % package_names)
        self._check_caller(sender)
        if self.role != ROLE_UNSET:
            raise TransactionRoleAlreadySet()
        action = policykit.PK_ACTION_UPGRADE_PACKAGES
        if not policykit.is_authorized_sender(action, sender):
            raise errors.NotAuthorizedError(action, sender)
        self.role = ROLE_UPGRADE_PACKAGES
        self.kwargs = {"package_names" : package_names}
        log_trans.debug("Queue transaction")
        self.queue.put(self)
        log_trans.debug("Queued")


    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="s", out_signature="",
                         sender_keyword="sender")
    def InstallFile(self, path, sender):
        """Install the given package file.

        Keyword arguments:
        path -- the absolute path to the package file
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("InstallFile() was called: %s" % path)
        self._check_caller(sender)
        if self.role != ROLE_UNSET:
            raise TransactionRoleAlreadySet()
        action = policykit.PK_ACTION_INSTALL_FILE
        if not policykit.is_authorized_sender(action, sender):
            raise errors.NotAuthorizedError(action, sender)
        #FIXME: Perform some checks
        self.role = ROLE_INSTALL_FILE
        self.kwargs = {"path" : path}
        log_trans.debug("Queue transaction")
        self.queue.put(self)
        log_trans.debug("Queued")

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="", out_signature="b")
    def GetAllowCancel(self):
        """Return True if the transaction can be cancelled."""
        log_trans.info("GetAllowCancel() was called")
        return self.allow_cancel

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="", out_signature="i")
    def GetRole(self):
        """Return the role of the transaction."""
        log_trans.info("GetRole() was called")
        return self.role

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="", out_signature="i")
    def GetStatus(self):
        """Return the current status of the transaction."""
        log_trans.info("GetStatus() was called")
        return self.status

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="", out_signature="s")
    def GetLocale(self):
        """Return the language and encoding of the transaction."""
        log_trans.info("GetLocale() was called")
        return self.locale

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="s", out_signature="",
                         sender_keyword="sender")
    def SetLocale(self, locale, sender):
        """Set the language and encoding.

        Keyword arguments:
        locale -- specifies language, territory and encoding according
                  to RFC 1766,  e.g. "de_DE.UTF-8"
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("SetLocale() was called: %s" % locale)
        self._check_caller(sender)
        if self.status != STATUS_SETTING_UP:
            raise TransactionAlreadyRunning()
        try:
            (lang, encoding) = parse_localename(locale)
        except ValueError:
            return None
        else:
            self.locale = "%s.%s" % (lang, encoding)

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="", out_signature="")
    def Attach(self):
        """Re-emit all signals."""
        self.Role(self.role)
        self.Status(self.status)
        self.Progress(self.progress)
        self.AllowCancel(self.allow_cancel)
        self.AllowTerminal(self.allow_terminal)
        if self.status_details:
            self.StatusDetails(self.status_details)
        for msg in self.messages:
            self.Message(msg.enum, msg.message)
        if self.required_medium:
            self.MediumRequired(self.required_medium[0], self.required_medium[1])

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="", out_signature="",
                         sender_keyword="sender")
    def Cancel(self, sender):
        """Cancel the transaction.

        Keyword argument:
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("Cancel() was called")
        try:
            self._check_caller(sender)
        except errors.ForeignTransaction:
            action = policykit.PK_ACTION_CANCEL_FOREIGN
            if not policykit.is_authorized_sender(action, sender):
                raise errors.NotAuthorizedError(action, sender)
        try:
            self.queue.remove(self)
            log_trans.debug("Removed transaction from queue")
        except ValueError:
            pass
        else:
            self.status = STATUS_CANCELLING
            self.exit = EXIT_CANCELLED
            return
        if self.allow_cancel:
            log_trans.debug("Setting cancel event")
            self.cancelled = True
            self.status = STATUS_CANCELLING
            self.paused = False
            return
        raise errors.APTDError("Could not cancel transaction")

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="s", out_signature="",
                         sender_keyword="sender")
    def SetTerminal(self, ttyname, sender):
        """Set the controlling terminal.

        The worker will be attached to the specified slave end of a pty
        master/slave pair. This allows to interact with the 

        Can only be changed before the transaction is started.

        Keyword arguments:
        ttyname -- file path to the slave file descriptor
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("SetTerminal() was called: %s" % ttyname)
        self._check_caller(sender)
        if self.status != STATUS_SETTING_UP:
            raise TransactionAlreadyRunning()
        if not os.access(ttyname, os.W_OK):
            raise errors.AptDError("Pty device does not exist: %s" % ttyname)
        if not os.stat(ttyname)[4] == self.caller.uid:
            raise errors.AptDError("Pty device '%s' has to be owned by the "
                                   "owner of the transaction " % ttyname)
        if self.status not in [STATUS_WAITING, STATUS_SETTING_UP]:
            raise errors.AptDError("You can only connect a terminal before the "
                                   "transaction gets processed.")
        try:
            slave_fd = os.open(ttyname, os.O_RDWR | os.O_NOCTTY)
            if os.isatty(slave_fd):
                self.terminal = ttyname
            else:
                raise errors.AptDError("%s isn't a tty" % ttyname)
        finally:
            os.close(slave_fd)
        return None

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="s", out_signature="",
                         sender_keyword="sender")
    def SetDebconf(self, debconf_socket, sender):
        """Set the socket of the debconf proxy.

        The worker process forwards all debconf commands through this
        socket by using the passthrough frontend. On the client side
        debconf-communicate should be connected to the socket.

        Can only be changed before the transaction is started.

        Keyword arguments:
        debconf_socket: absolute path to the socket
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("SetDebconf() was called: %s" % debconf_socket)
        self._check_caller(sender)
        if self.status != STATUS_SETTING_UP:
            raise TransactionAlreadyRunning()
        if not os.access(debconf_socket, os.W_OK):
            raise errors.AptDError("socket does not exist: %s" % debconf_socket)
        if not os.stat(debconf_socket)[4] == self.caller.uid:
            raise errors.AptDError("socket '%s' has to be owned by the "
                                   "owner of the transaction " % ttyname)
        if self.status not in [STATUS_WAITING, STATUS_SETTING_UP]:
            raise errors.AptDError("You can only connect debconf socket before "
                                   "the transaction gets processed.")
        self.debconf = debconf_socket
        return None

    @dbus.service.method(APTDAEMON_TRANSACTION_DBUS_INTERFACE,
                         in_signature="s", out_signature="",
                         sender_keyword="sender")
    def ProvideMedium(self, medium, sender):
        """Continue paused transaction with the inserted medium.

        If a media change is required to install packages from CD/DVD
        the transaction will be paused and could be resumed with this
        method.

        Keyword arguments:
        medium -- the label of the CD/DVD
        sender -- the unique D-Bus name of the sender (provided by D-Bus)
        """
        log_trans.info("ProvideMedium() was called: %s" % medium)
        self._check_caller(sender)
        if not self.required_medium:
            raise AptDError("There isn't any required medium.")
        if not self.required_medium[0] == medium:
            raise AptDError("The medium '%s' was not requested." % medium)
        self.paused = False

    def send_message(self, enum, msg):
        """Send a non-critical message to the user.

        Keyword arguments:
        enum -- the message type
        msg -- free text information
        """
        self.messages.append(Message(enum, msg))
        self.Message(enum, msg)

    def _check_caller(self, dbus_name):
        """Check if the given dbus name belongs to the user who intiated the
        transaction.
        """
        caller = policykit.get_caller_from_dbus_name(dbus_name)
        if caller.uid != self.caller.uid or \
           caller.session != self.caller.session:
            raise errors.ForeignTransaction()


class TransactionQueue(gobject.GObject):

    """Queue for transactions."""

    __gsignals__ = {"queue-changed":(gobject.SIGNAL_RUN_FIRST,
                                     gobject.TYPE_NONE,
                                     ())}

    def __init__(self):
        """Intialize a new TransactionQueue instance."""
        gobject.GObject.__init__(self)
        self._queue = collections.deque()
        self._proc_count = 0
        self.worker = AptWorker()
        self.worker.connect("transaction-done", self._on_transaction_done)

    def __len__(self):
        return len(self._queue)

    def _emit_queue_changed(self):
        """Emit the queued-changed signal."""
        log.debug("emitting queue changed")
        self.emit("queue-changed")

    def put(self, transaction):
        """Add an item to the queue."""
        if self.worker.trans:
            self._queue.append(transaction)
        else:
            self.worker.run(transaction)
        self._emit_queue_changed()

    def _on_transaction_done(self, worker, tid):
        """Mark the last item as done and request a new item."""
        try:
            next = self._queue.popleft()
        except IndexError:
            log.debug("There isn't any queued transaction")
        else:
            self.worker.run(next)
        self._emit_queue_changed()

    def remove(self, transaction):
        """Remove the specified item from the queue."""
        self._queue.remove(transaction)
        self._emit_queue_changed()

    def clear(self):
        """Remove all items from the queue."""
        self._queue.clear()

    @property
    def items(self):
        """Return a list containing all queued items."""
        return list(self._queue)


class AptDaemon(dbus.service.Object):

    """Provides a system daemon to process package management tasks.

    The daemon is transaction based. Each package management tasks runs
    in a separate transaction. The transactions can be created,
    monitored and managed via the D-Bus interface.
    """

    def __init__(self, options, connect=True, bus=None):
        """Initialize a new AptDaemon instance.

        Keyword arguments:
        options -- command line options of the type optparse.Values
        connect -- if the daemon should connect to the D-Bus (default is True)
        bus -- the D-Bus to connect to (defaults to the system bus)
        """
        log.info("Initializing daemon")
        signal.signal(signal.SIGQUIT, self._sigquit)
        signal.signal(signal.SIGTERM, self._sigquit)
        self.options = options
        if connect == True:
            if bus is None:
                bus = dbus.SystemBus()
            # Check if another object has already registered the name on
            # the bus. Quit the other daemon if replace would be set
            try:
                bus_name = dbus.service.BusName(APTDAEMON_DBUS_SERVICE,
                                                bus,
                                                do_not_queue=True)
            except dbus.exceptions.NameExistsException:
                if self.options.replace == False:
                    log.critical("Another daemon is already running")
                    sys.exit(1)
                log.warn("Replacing already running daemon")
                the_other_guy = bus.get_object("org.debian.apt",
                                               "/org/debian/apt")
                the_other_guy.Quit(dbus_interface="org.debian.apt",
                                   timeout=300)
                bus_name = dbus.service.BusName(APTDAEMON_DBUS_SERVICE,
                                                bus,
                                                do_not_queue=True)
        else:
            bus_name = None
        dbus.service.Object.__init__(self, bus_name, APTDAEMON_DBUS_PATH)
        self.loop = get_main_loop()
        self.queue = TransactionQueue()
        self.queue.connect("queue-changed", self._on_queue_changed)
        log.debug("Setting up worker thread")
        log.debug("Daemon was initialized")

    def _on_queue_changed(self, queue):
        """Callback for a changed transaction queue."""
        if self.queue.worker.trans:
            current = self.queue.worker.trans.tid
        else:
            current = ""
        queued = [trans.tid for trans in self.queue.items]
        self.ActiveTransactionsChanged(current, queued)

    @dbus.service.signal(dbus_interface=APTDAEMON_DBUS_INTERFACE,
                         signature="sas")
    def ActiveTransactionsChanged(self, current, queued):
        """Emit the ActiveTransactionsChanged signal.

        Keyword arguments:
        current -- the tid of the currently running transaction or an empty
                   string
        queued -- list of the ids of the queued transactions
        """
        log.debug("Emitting ActiveTransactionsChanged signal: %s, %s",
                  current, queued)
 
    def run(self):
        """Start the daemon and listen for calls."""
        if self.options.disable_timeout == False:
            log.debug("Using inactivity check")
            gobject.timeout_add_seconds(INACTIVITY_CHECK_INTERVAL,
                                        self._check_for_inactivity)
        log.debug("Waiting for calls")
        try:
            self.loop.run()
        except KeyboardInterrupt:
            self.Quit(None)

    @dbus.service.method(APTDAEMON_DBUS_INTERFACE,
                         in_signature="", out_signature="s",
                         sender_keyword="caller_name")
    def CreateTransaction(self, caller_name):
        """Create a new transaction and return its unique identifier.

        The transaction id is the D-Bus path of the transaction.

        Keyword argument:
        caller_name -- the D-Bus name of the caller (provided by D-Bus)
        """
        log.debug("Creating transaction")
        caller = policykit.get_caller_from_dbus_name(caller_name)
        trans = Transaction(queue=self.queue, caller=caller)
        return trans.tid

    @dbus.service.method(APTDAEMON_DBUS_INTERFACE,
                         in_signature="", out_signature="sas")
    def GetActiveTransactions(self):
        """Return the currently running transaction and the list of queued
        transactions.
        """
        log.debug("GetActiveTransactions() was called")
        queued = [trans.tid for trans in self.queue.items]
        if self.queue.worker.trans:
            current = self.queue.worker.trans.tid
        else:
            current = ""
        return current, queued

    @dbus.service.method(APTDAEMON_DBUS_INTERFACE,
                         in_signature="", out_signature="",
                         sender_keyword="caller_name")
    def Quit(self, caller_name):
        """Request a shutdown of the daemon.

        Keyword argument:
        caller_name -- the D-Bus name of the caller (provided by D-Bus)
        """
        log.critical("Shutdown was requested")
        log.debug("Quitting main loop...")
        self.loop.quit()
        log.debug("Exit")

    def _sigquit(self, signum, frame):
        """Internal callback for the quit signal."""
        self.Quit(None)

    def _check_for_inactivity(self):
        """Shutdown the daemon if it has been inactive for time specified
        in INACTIVITY_TIMEOUT.
        """
        log.debug("Checking for inactivity")
        timestamp = self.queue.worker.last_action_timestamp
        if not self.worker.trans and \
           not self.loop.get_context().pending() and \
           time.time() - timestamp > INACTIVITY_TIMEOUT and \
           not self.queue:
            log.info("Quiting due to inactivity")
            self.Quit(None)
            return False
        return True


def main():
    """Allow to run the daemon from the command line."""
    parser = OptionParser()
    parser.add_option("-t", "--disable-timeout",
                      default=False,
                      action="store_true", dest="disable_timeout",
                      help=_("Do not shutdown the daemon because of "
                             "inactivity"))
    parser.add_option("-d", "--debug",
                      default=False,
                      action="store_true", dest="debug",
                      help=_("Show internal processing information"))
    parser.add_option("-r", "--replace",
                      default=False,
                      action="store_true", dest="replace",
                      help=_("Quit and replace an already running daemon"))
    parser.add_option("-p", "--profile",
                      default=False,
                      action="store", type="string", dest="profile",
                      help=_("Store profile stats in the specified file"))
    options, args = parser.parse_args()
    if options.debug == True:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)
        _syslog_handler.setLevel(logging.WARN)
        _console_handler.setLevel(logging.INFO)
    daemon = AptDaemon(options)
    if options.profile:
        import profile
        profiler = profile.Profile()
        profiler.runcall(daemon.run)
        profiler.dump_stats(options.profile)
        profiler.print_stats()
    else:
        daemon.run()

# vim:ts=4:sw=4:et
