#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Manage privillegs using PolicyKit."""
# Copyright (C) 2008-2009 Sebastian Heinlein <devel@glatzor.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

__author__  = "Sebastian Heinlein <devel@glatzor.de>"

import os

import dbus

PK_ACTION_REMOVE_PACKAGES = "org.debian.apt.remove-packages"
PK_ACTION_INSTALL_PACKAGES = "org.debian.apt.install-packages"
PK_ACTION_INSTALL_FILE = "org.debian.apt.install-file"
PK_ACTION_UPGRADE_PACKAGES = "org.debian.apt.upgrade-packages"
PK_ACTION_UPDATE_CACHE = "org.debian.apt.update-cache"
PK_ACTION_UPGRADE_SYSTEM = "org.debian.apt.upgrade-system"
PK_ACTION_CANCEL_FOREIGN = "org.debian.apt.cancel-foreign"

_CALLER_CACHE = {}

class Caller:

    """Container for information about the caller of a D-Bus method."""

    def __init__(self, name):
        """Initialze a new Caller instance.

        Keyword arguments:
        name -- the unique D-Bus name of the caller
        """
        self.uid = None
        self.pid = None
        self.session = None
        self.name = name

    def __str__(self):
        return "Caller uid: %s, pid: %s, session: %s, name: %s" % \
                (self.uid, self.pid, self.session, self.name)

def get_caller_from_dbus_name(dbus_name):
    """Return a Caller instance for the given D-Bus caller name.

    The Caller instance provides the user and process id and the
    console kit session.

    This is a partial Python re-implementation of
    polkit_caller_new_from_dbus_name from libpolkit-dbus2
    """
    if _CALLER_CACHE.has_key(dbus_name):
        return _CALLER_CACHE[dbus_name]
    caller = Caller(dbus_name)
    # Get the pid from dbus
    bus = dbus.SystemBus()
    bus_proxy = bus.get_object("org.freedesktop.DBus",
                               "/org/freedesktop/DBus/Bus")
    bus_iface = dbus.Interface(bus_proxy, "org.freedesktop.DBus")
    caller.pid = bus_iface.GetConnectionUnixProcessID(dbus_name)
    # Get user id from /proc
    proc = open("/proc/%s/status" % caller.pid)
    values = [v for v in proc.readlines() if v.startswith("Uid:")]
    proc.close()
    caller.uid = int(values[0].split()[1])
    # Get consolekit session
    ck_proxy = bus.get_object("org.freedesktop.ConsoleKit",
                               "/org/freedesktop/ConsoleKit/Manager")
    ck_iface = dbus.Interface(ck_proxy, "org.freedesktop.ConsoleKit.Manager")
    caller.session = ck_iface.GetSessionForUnixProcess(caller.pid)
    return caller

def is_authorized(action):
    """Check if the current process is authorized for the given action.

    Keyword argument:
    action -- the PolicyKit policy name of the action
    """
    bus = dbus.SystemBus()
    pk = bus.get_object("org.freedesktop.PolicyKit", "/")
    pk_iface = dbus.Interface(pk, "org.freedesktop.PolicyKit")
    pid = os.getpid()
    if not pk_iface.IsProcessAuthorized(action, pid, False,
                                        timeout=300) == "yes":
        return False
    return True

def is_authorized_sender(action, sender):
    """Check if the given sender is authorized for the specified action.

    Keyword arguments:
    sender -- D-Bus name of the sender
    action -- the PolicyKit policy name of the action
    """
    bus = dbus.SystemBus()
    pk = bus.get_object("org.freedesktop.PolicyKit", "/")
    pk_iface = dbus.Interface(pk, "org.freedesktop.PolicyKit")
    if not pk_iface.IsSystemBusNameAuthorized(action, sender, False,
                                              timeout=300) == "yes":
        return False
    return True

def obtain_authorization(action, xid=0):
    """Try to obtain the authorization for the specified action.

    Calls the session based authentication agent which provides a dialog
    for authentication.

    Keyword arguments:
    action -- the PolicyKit policy name of the action
    xid -- id of the window for which the auth dialog should be transient
    """
    pid = os.getpid()
    bus = dbus.SessionBus()
    iface="org.freedesktop.PolicyKit.AuthenticationAgent"
    aa = bus.get_object("org.freedesktop.PolicyKit.AuthenticationAgent", "/")
    return aa.ObtainAuthorization(action, xid, pid,
                                  timeout=300, dbus_interface=iface)

# vim:ts=4:sw=4:et
